/*-
 * Copyright (c) 2010 Hans Petter Selasky. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/stdint.h>
#include <sys/stddef.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/linker_set.h>
#include <sys/module.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/condvar.h>
#include <sys/sysctl.h>
#include <sys/unistd.h>
#include <sys/malloc.h>
#include <sys/priv.h>
#include <sys/uio.h>
#include <sys/poll.h>
#include <sys/sx.h>

#include <machine/bus.h>

#include <vm/vm.h>
#include <vm/pmap.h>

#include "video4bsd.h"

MODULE_VERSION(video4bsd, 1);

struct v4b_memory {
	uint8_t *virtaddr;
	uint32_t page_count;
	uint8_t	alloc_done;
	uint8_t	in_use_by_daemon;
	uint8_t	in_use_by_dev;
};

struct v4b_softc {
	struct mtx sc_mtx;
	struct sx sc_sx;
	struct cv sc_cv;
	struct v4b_memory sc_mem[V4B_ALLOC_UNIT_MAX];

	struct cdev *sc_dev;
	struct cdev *sc_daemon;

	uint32_t sc_daemon_len;

	uint32_t sc_mmap_page_offset;

	uint32_t sc_buffer[V4B_BUFFER_MAX / 4];

	struct v4b_command sc_cmd_to_daemon;
	struct v4b_data_chunk sc_daemon_chunk;

	int	sc_error_to_daemon;
	int	sc_error_from_daemon;

	uint8_t	sc_cmd_from_daemon;

	uint8_t	sc_daemon_ready;
	uint8_t	sc_daemon_opened;
	uint8_t	sc_dev_opened;
	uint8_t	sc_mmap_alloc_nr;
};

static struct v4b_softc v4b_softc[V4B_DEVICES_MAX];

static d_open_t v4b_dev_open;
static d_close_t v4b_dev_close;
static d_ioctl_t v4b_dev_ioctl;
static d_read_t v4b_dev_read;
static d_write_t v4b_dev_write;
static d_poll_t v4b_dev_poll;
static d_mmap_t v4b_dev_mmap;

static struct cdevsw v4b_dev_devsw = {
	.d_version = D_VERSION,
	.d_open = v4b_dev_open,
	.d_close = v4b_dev_close,
	.d_ioctl = v4b_dev_ioctl,
	.d_name = "v4b_dev",
	.d_flags = D_TRACKCLOSE,
	.d_read = v4b_dev_read,
	.d_write = v4b_dev_write,
	.d_poll = v4b_dev_poll,
	.d_mmap = v4b_dev_mmap,
};

static d_open_t v4b_daemon_open;
static d_close_t v4b_daemon_close;
static d_ioctl_t v4b_daemon_ioctl;
static d_read_t v4b_daemon_read;
static d_write_t v4b_daemon_write;
static d_poll_t v4b_daemon_poll;
static d_mmap_t v4b_daemon_mmap;

static struct cdevsw v4b_daemon_devsw = {
	.d_version = D_VERSION,
	.d_open = v4b_daemon_open,
	.d_close = v4b_daemon_close,
	.d_ioctl = v4b_daemon_ioctl,
	.d_name = "v4b_daemon",
	.d_flags = D_TRACKCLOSE,
	.d_read = v4b_daemon_read,
	.d_write = v4b_daemon_write,
	.d_poll = v4b_daemon_poll,
	.d_mmap = v4b_daemon_mmap,
};

static int v4b_free_memory(struct v4b_softc *, struct v4b_memory *, uint8_t);
static void v4b_lock(struct v4b_softc *, int);
static void v4b_unlock(struct v4b_softc *, int);

static void
v4b_init(void *arg)
{
	struct v4b_softc *sc;
	uint8_t i;

	for (i = 0; i != V4B_DEVICES_MAX; i++) {
		sc = &v4b_softc[i];

		mtx_init(&sc->sc_mtx, "video4bsd", NULL, MTX_DEF);
		sx_init(&sc->sc_sx, "video4bsd");
		cv_init(&sc->sc_cv, "V4B");

		sc->sc_dev = make_dev(&v4b_dev_devsw, 0,
		    UID_ROOT, GID_OPERATOR, 0644, "video%d", i);

		if (sc->sc_dev != NULL)
			sc->sc_dev->si_drv1 = sc;

		sc->sc_daemon = make_dev(&v4b_daemon_devsw, 0,
		    UID_ROOT, GID_OPERATOR, 0600, "video_daemon%d", i);

		if (sc->sc_daemon != NULL)
			sc->sc_daemon->si_drv1 = sc;
	}

	printf("video4bsd: /dev/video0..%d, /dev/video_daemon0..%d\n",
	    V4B_DEVICES_MAX - 1, V4B_DEVICES_MAX - 1);
}

SYSINIT(v4b_init, SI_SUB_DEVFS, SI_ORDER_FIRST, v4b_init, 0);

static void
v4b_uninit(void *arg)
{
	struct v4b_softc *sc;
	uint8_t i;
	uint8_t n;

	for (i = 0; i != V4B_DEVICES_MAX; i++) {
		sc = &v4b_softc[i];

		while (sc->sc_daemon_ready || sc->sc_dev_opened) {
			/* In use */
			printf("video4bsd: Please exit any video device "
			    "clients for device #%d.\n", i);
			pause("DRAIN", hz);
		}

		if (sc->sc_dev != NULL)
			destroy_dev(sc->sc_dev);

		if (sc->sc_daemon != NULL)
			destroy_dev(sc->sc_daemon);

		v4b_lock(sc, 0);

		for (n = 0; n != V4B_ALLOC_UNIT_MAX; n++) {
			v4b_free_memory(sc, &sc->sc_mem[n], 0);
			v4b_free_memory(sc, &sc->sc_mem[n], 1);
		}

		v4b_unlock(sc, 0);

		mtx_destroy(&sc->sc_mtx);
		sx_destroy(&sc->sc_sx);
		cv_destroy(&sc->sc_cv);
	}
}

SYSUNINIT(v4b_uninit, SI_SUB_DEVFS, SI_ORDER_ANY, v4b_uninit, 0);

static void
v4b_lock(struct v4b_softc *sc, int lock_sx)
{
	if (lock_sx)
		sx_xlock(&sc->sc_sx);
	mtx_lock(&sc->sc_mtx);
}

static void
v4b_unlock(struct v4b_softc *sc, int unlock_sx)
{
	mtx_unlock(&sc->sc_mtx);
	if (unlock_sx)
		sx_xunlock(&sc->sc_sx);
}

static int
v4b_set_command(struct v4b_softc *sc, int cmd, int arg, void *ptr)
{
	int error = 0;

	while (sc->sc_cmd_to_daemon.command != V4B_CMD_NONE) {
		error = cv_wait_sig(&sc->sc_cv, &sc->sc_mtx);
		if (error)
			break;
		if (sc->sc_daemon_ready == 0)
			return (ENXIO);

	}

	if (error)
		goto done;

	sc->sc_cmd_to_daemon.command = cmd;
	sc->sc_cmd_to_daemon.arg = arg;
	sc->sc_cmd_to_daemon.ptr = ptr;
	cv_signal(&sc->sc_cv);

done:
	return (error);
}

static int
v4b_alloc_memory(struct v4b_softc *sc,
    struct v4b_memory *mem, uint32_t page_count)
{
	void *ptr;
	int err;

	if (mem->virtaddr != NULL) {
		if (mem->in_use_by_daemon) {
			return (EBUSY);
		}
		if (mem->page_count == page_count) {
			mem->in_use_by_daemon = 1;
			return (0);
		}
		return (EBUSY);
	}
	memset(mem, 0, sizeof(*mem));

	mem->virtaddr = (uint8_t *)1;

	v4b_unlock(sc, 0);

	ptr = malloc(page_count * PAGE_SIZE, M_TEMP, M_WAITOK | M_ZERO);
	if (ptr == NULL)
		err = ENOMEM;
	else
		err = 0;

	v4b_lock(sc, 0);

	if (err) {
		mem->virtaddr = NULL;
		return (err);
	}
	mem->virtaddr = ptr;
	mem->page_count = page_count;
	mem->in_use_by_daemon = 1;
	return (0);
}

static int
v4b_free_memory(struct v4b_softc *sc, struct v4b_memory *mem, uint8_t is_daemon)
{
	if (mem->virtaddr == NULL)
		return (ENXIO);

	if (is_daemon)
		mem->in_use_by_daemon = 0;
	else
		mem->in_use_by_dev = 0;

	if (mem->in_use_by_daemon ||
	    mem->in_use_by_dev)
		return (EBUSY);

	v4b_unlock(sc, 0);

	free(mem->virtaddr, M_TEMP);

	v4b_lock(sc, 0);

	mem->virtaddr = NULL;

	return (0);
}

static int
v4b_io_loop(struct v4b_softc *sc, struct uio *uio,
    uint8_t *arg_ptr, uint32_t arg_len, int client_cmd)
{
	int complete = 0;
	int error = 0;
	int daemon_cmd;
	uint32_t offset;

	if (sc->sc_daemon_ready == 0)
		return (ENXIO);

	while (!complete) {

		while (sc->sc_cmd_from_daemon == V4B_CMD_NONE) {
			error = cv_wait_sig(&sc->sc_cv, &sc->sc_mtx);
			if (error)
				break;
			if (sc->sc_daemon_ready == 0)
				return (ENXIO);
		}

		if (error)
			return (error);

		daemon_cmd = sc->sc_cmd_from_daemon;

		v4b_unlock(sc, 0);

		switch (daemon_cmd) {
		case V4B_CMD_READ:
#ifdef V4B_DEBUG
			printf("copyin(%p,%p,%d)\n",
			    sc->sc_daemon_chunk.peer_ptr,
			    sc->sc_buffer,
			    sc->sc_daemon_len);
#endif
			switch (client_cmd) {
			case V4B_CMD_WRITE:
				error = uiomove(sc->sc_buffer, sc->sc_daemon_len, uio);
				break;
			case V4B_CMD_IOCTL:
				if (sc->sc_daemon_chunk.peer_ptr < V4B_BUF_MIN_PTR) {
					error = EINVAL;
				} else if (sc->sc_daemon_chunk.peer_ptr < V4B_BUF_MAX_PTR) {
					offset = sc->sc_daemon_chunk.peer_ptr - V4B_BUF_MIN_PTR;
					if (offset >= arg_len) {
						error = EINVAL;
						break;
					}
					if ((uint32_t)(offset + sc->sc_daemon_len) > arg_len) {
						error = EINVAL;
						break;
					}
					if ((uint32_t)(offset + sc->sc_daemon_len) < offset) {
						error = EINVAL;
						break;
					}
					memcpy(sc->sc_buffer, arg_ptr + offset, sc->sc_daemon_len);
					error = 0;
				} else {
					error = copyin(
					    sc->sc_daemon_chunk.peer_ptr,
					    sc->sc_buffer,
					    sc->sc_daemon_len);
				}
				break;
			default:
				error = EINVAL;
				break;
			}
			break;

		case V4B_CMD_WRITE:
#ifdef V4B_DEBUG
			printf("copyout(%p,%p,%d,%d)\n",
			    sc->sc_buffer,
			    sc->sc_daemon_chunk.peer_ptr,
			    sc->sc_daemon_len, arg_len);
#endif

			switch (client_cmd) {
			case V4B_CMD_READ:
				error = uiomove(sc->sc_buffer, sc->sc_daemon_len, uio);
				break;
			case V4B_CMD_IOCTL:
				if (sc->sc_daemon_chunk.peer_ptr < V4B_BUF_MIN_PTR) {
					error = EINVAL;
				} else if (sc->sc_daemon_chunk.peer_ptr < V4B_BUF_MAX_PTR) {
					offset = sc->sc_daemon_chunk.peer_ptr - V4B_BUF_MIN_PTR;
					if (offset >= arg_len) {
						error = EINVAL;
						break;
					}
					if ((uint32_t)(offset + sc->sc_daemon_len) > arg_len) {
						error = EINVAL;
						break;
					}
					if ((uint32_t)(offset + sc->sc_daemon_len) < offset) {
						error = EINVAL;
						break;
					}
					memcpy(arg_ptr + offset, sc->sc_buffer, sc->sc_daemon_len);
					error = 0;
				} else {
					error = copyout(
					    sc->sc_buffer,
					    sc->sc_daemon_chunk.peer_ptr,
					    sc->sc_daemon_len);
				}
				break;
			default:
				error = EINVAL;
				break;
			}
			break;

		case V4B_CMD_SYNC:
#ifdef V4B_DEBUG
			printf("Sync received %d\n", sc->sc_error_from_daemon);
#endif
			error = sc->sc_error_from_daemon;
			if (error)
				error = ENXIO;
			complete = 1;
			break;

		default:
			error = EINVAL;
			break;
		}

		v4b_lock(sc, 0);

		sc->sc_error_to_daemon = error;
		sc->sc_cmd_from_daemon = V4B_CMD_NONE;
		cv_signal(&sc->sc_cv);
	}

	return (error);
}

/*------------------------------------------------------------------------*
 *	VIDEO4BSD DEVICE PART
 *------------------------------------------------------------------------*/

static int
v4b_dev_open(struct cdev *dev, int fflags, int devtype, struct thread *td)
{
	struct v4b_softc *sc = (struct v4b_softc *)dev->si_drv1;
	int error = 0;

	if (sc == NULL)
		return (ENXIO);

	v4b_lock(sc, 1);

	if (sc->sc_dev_opened != 0) {
		v4b_unlock(sc, 1);
		return (EBUSY);
	}
	if (sc->sc_daemon_ready == 0) {
		v4b_unlock(sc, 1);
		return (ENXIO);
	}
	sc->sc_dev_opened = 1;
	sc->sc_cmd_from_daemon = V4B_CMD_NONE;

	error = v4b_set_command(sc, V4B_CMD_OPEN, 0, 0);

	if (error == 0)
		error = v4b_io_loop(sc, 0, 0, 0, V4B_CMD_OPEN);

	if (error == 0) {
		sc->sc_daemon_opened = 1;
	} else {
		sc->sc_dev_opened = 0;
	}

	v4b_unlock(sc, 1);

	return (error);
}

static int
v4b_dev_close(struct cdev *dev, int fflag, int devtype, struct thread *td)
{
	struct v4b_softc *sc = (struct v4b_softc *)dev->si_drv1;
	int error;
	uint8_t n;

	v4b_lock(sc, 1);

	error = v4b_set_command(sc, V4B_CMD_CLOSE, 0, 0);

	if (error == 0)
		v4b_io_loop(sc, 0, 0, 0, V4B_CMD_CLOSE);

	for (n = 0; n != V4B_ALLOC_UNIT_MAX; n++)
		v4b_free_memory(sc, &sc->sc_mem[n], 0);

	/* reset state */
	sc->sc_cmd_to_daemon.command = V4B_CMD_NONE;
	sc->sc_daemon_opened = 0;
	sc->sc_dev_opened = 0;
	cv_signal(&sc->sc_cv);

	v4b_unlock(sc, 1);

	return (0);
}

static int
v4b_dev_read(struct cdev *dev, struct uio *uio, int ioflag)
{
	struct v4b_softc *sc = (struct v4b_softc *)dev->si_drv1;
	int resid = uio->uio_resid;
	int error = 0;

	if (resid < 0)
		return (EINVAL);

	v4b_lock(sc, 1);

	if (sc->sc_daemon_opened == 0) {
		v4b_unlock(sc, 1);
		return (ENXIO);
	}
	error = v4b_set_command(sc, V4B_CMD_READ, resid, V4B_BUF_MIN_PTR);

	if (error == 0)
		error = v4b_io_loop(sc, uio, 0, 0, V4B_CMD_READ);

	v4b_unlock(sc, 1);

	return (error);
}

static int
v4b_dev_write(struct cdev *dev, struct uio *uio, int ioflag)
{
	struct v4b_softc *sc = (struct v4b_softc *)dev->si_drv1;
	int resid = uio->uio_resid;
	int error = 0;

	if (resid < 0)
		return (EINVAL);

	v4b_lock(sc, 1);

	if (sc->sc_daemon_opened == 0) {
		v4b_unlock(sc, 1);
		return (ENXIO);
	}
	error = v4b_set_command(sc, V4B_CMD_WRITE, resid, V4B_BUF_MIN_PTR);

	if (error == 0)
		error = v4b_io_loop(sc, uio, 0, 0, V4B_CMD_WRITE);

	v4b_unlock(sc, 1);

	return (error);
}

int
v4b_dev_ioctl(struct cdev *dev, u_long cmd, caddr_t data, int fflag,
    struct thread *td)
{
	struct v4b_softc *sc = (struct v4b_softc *)dev->si_drv1;
	int error;
	int len;

	v4b_lock(sc, 1);

	if (sc->sc_daemon_opened == 0) {
		v4b_unlock(sc, 1);
		return (ENXIO);
	}
	len = IOCPARM_LEN(cmd);
	if (len > V4B_BUFFER_MAX) {
		v4b_unlock(sc, 1);
		return (ENOMEM);
	}
	error = v4b_set_command(sc, V4B_CMD_IOCTL, cmd, V4B_BUF_MIN_PTR);

	if (error == 0)
		error = v4b_io_loop(sc, 0, data, len, V4B_CMD_IOCTL);

	v4b_unlock(sc, 1);

	return (error);
}

static int
v4b_dev_poll(struct cdev *dev, int events, struct thread *td)
{
	return (events & (POLLHUP | POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM));
}

static int
#if (__FreeBSD_version >= 900000)
v4b_dev_mmap(struct cdev *dev, vm_ooffset_t offset, vm_paddr_t *paddr, int nprot, vm_memattr_t *memattr)
#else
v4b_dev_mmap(struct cdev *dev, vm_offset_t offset, vm_paddr_t *paddr, int nprot)
#endif
{
	struct v4b_softc *sc = (struct v4b_softc *)dev->si_drv1;
	int error;
	void *ptr;
	uint32_t alloc_nr;
	uint32_t page_nr;

	v4b_lock(sc, 1);

	if (sc->sc_daemon_opened == 0) {
		v4b_unlock(sc, 1);
		return (ENXIO);
	}
	sc->sc_mmap_alloc_nr = 0 - 1;

	error = v4b_set_command(sc, V4B_CMD_MMAP, offset, 0);

	if (error == 0)
		error = v4b_io_loop(sc, 0, 0, 0, V4B_CMD_MMAP);

	if (error == 0) {
		alloc_nr = sc->sc_mmap_alloc_nr;
		page_nr = sc->sc_mmap_page_offset;

		if ((alloc_nr >= V4B_ALLOC_UNIT_MAX) ||
		    (sc->sc_mem[alloc_nr].virtaddr < (uint8_t *)2) ||
		    (sc->sc_mem[alloc_nr].page_count <= page_nr)) {
			error = ENXIO;
		} else {

			ptr = sc->sc_mem[alloc_nr].virtaddr + (page_nr * PAGE_SIZE);
			sc->sc_mem[alloc_nr].in_use_by_dev = 1;

			v4b_unlock(sc, 0);
			*paddr = vtophys(ptr);
			v4b_lock(sc, 0);
		}
	}
	v4b_unlock(sc, 1);

	return (error);
}

/*------------------------------------------------------------------------*
 *	VIDEO4BSD DAEMON PART
 *------------------------------------------------------------------------*/

static int
v4b_daemon_open(struct cdev *dev, int fflags, int devtype, struct thread *td)
{
	struct v4b_softc *sc = (struct v4b_softc *)dev->si_drv1;

	if (sc == NULL)
		return (ENXIO);

	v4b_lock(sc, 0);

	if (sc->sc_daemon_ready != 0) {
		v4b_unlock(sc, 0);
		return (ENXIO);
	}
	sc->sc_daemon_ready = 1;

	v4b_unlock(sc, 0);

	return (0);
}

static int
v4b_daemon_close(struct cdev *dev, int fflag, int devtype, struct thread *td)
{
	struct v4b_softc *sc = (struct v4b_softc *)dev->si_drv1;
	uint8_t n;

	v4b_lock(sc, 0);

	for (n = 0; n != V4B_ALLOC_UNIT_MAX; n++)
		v4b_free_memory(sc, &sc->sc_mem[n], 1);

	sc->sc_daemon_ready = 0;
	sc->sc_daemon_opened = 0;
	cv_signal(&sc->sc_cv);

	v4b_unlock(sc, 0);

	return (0);
}

static int
v4b_daemon_read(struct cdev *dev, struct uio *uio, int ioflag)
{
	return (ENXIO);
}

static int
v4b_daemon_write(struct cdev *dev, struct uio *uio, int ioflag)
{
	return (ENXIO);
}

int
v4b_daemon_ioctl(struct cdev *dev, u_long cmd, caddr_t data, int fflag,
    struct thread *td)
{
	struct v4b_softc *sc = (struct v4b_softc *)dev->si_drv1;
	struct v4b_mem_alloc *mac;
	int error = 0;
	int first = 1;
	uint32_t temp;

	v4b_lock(sc, 0);

	switch (cmd) {
	case V4B_IOCTL_GET_COMMAND:
		while (sc->sc_cmd_to_daemon.command == V4B_CMD_NONE) {
			error = cv_wait_sig(&sc->sc_cv, &sc->sc_mtx);
			if (error) {
				v4b_unlock(sc, 0);
				return (error);
			}
		}

		*(struct v4b_command *)data = sc->sc_cmd_to_daemon;

		/* clear command */
		sc->sc_cmd_to_daemon.command = V4B_CMD_NONE;
		cv_signal(&sc->sc_cv);
		break;

	case V4B_IOCTL_SYNC_COMMAND:
		/* wait for ready state */

#ifdef V4B_DEBUG
		printf("Sync transmitted (1)\n");
#endif
		while (sc->sc_cmd_from_daemon != V4B_CMD_NONE) {

			error = cv_wait_sig(&sc->sc_cv, &sc->sc_mtx);
			if (error) {
				v4b_unlock(sc, 0);
				return (error);
			}
		}

#ifdef V4B_DEBUG
		printf("Sync transmitted (2)\n");
#endif
		sc->sc_cmd_from_daemon = V4B_CMD_SYNC;
		sc->sc_error_from_daemon = *(int *)data;
		cv_signal(&sc->sc_cv);
		break;

	case V4B_IOCTL_WRITE_DATA:

		if (sc->sc_daemon_opened == 0) {
			error = ENXIO;
			break;
		}
		temp = 0;

		while (1) {

			/* wait for ready state */

			while (sc->sc_cmd_from_daemon != V4B_CMD_NONE) {

				error = cv_wait_sig(&sc->sc_cv, &sc->sc_mtx);
				if (sc->sc_daemon_opened == 0) {
					v4b_unlock(sc, 0);
					return (ENXIO);
				}
				if (error) {
					v4b_unlock(sc, 0);
					return (error);
				}
			}

			if (first) {
				sc->sc_daemon_chunk = *(struct v4b_data_chunk *)data;
				sc->sc_error_to_daemon = 0;
				first = 0;
			}
			error = sc->sc_error_to_daemon;
			if (error)
				break;

			sc->sc_daemon_chunk.peer_ptr += temp;

			/* check if complete */
			if (sc->sc_daemon_chunk.length == 0)
				break;

			temp = sc->sc_daemon_chunk.length;
			if (temp > V4B_BUFFER_MAX)
				temp = V4B_BUFFER_MAX;

			/* transfer data */
			v4b_unlock(sc, 0);
#ifdef V4B_DEBUG
			printf("copyinB(%p,%p,%d)\n",
			    sc->sc_daemon_chunk.client_ptr,
			    sc->sc_buffer,
			    temp);
#endif
			error = copyin(sc->sc_daemon_chunk.client_ptr, sc->sc_buffer, temp);
			v4b_lock(sc, 0);

			if (error)
				break;

			sc->sc_daemon_chunk.client_ptr += temp;
			sc->sc_daemon_chunk.length -= temp;
			sc->sc_daemon_len = temp;
			sc->sc_cmd_from_daemon = V4B_CMD_WRITE;
			cv_signal(&sc->sc_cv);
		}
		break;

	case V4B_IOCTL_READ_DATA:

		if (sc->sc_daemon_opened == 0) {
			error = ENXIO;
			break;
		}
		while (1) {

			/* wait for ready state */

			while (sc->sc_cmd_from_daemon != V4B_CMD_NONE) {

				error = cv_wait_sig(&sc->sc_cv, &sc->sc_mtx);
				if (sc->sc_daemon_opened == 0) {
					v4b_unlock(sc, 0);
					return (ENXIO);
				}
				if (error) {
					v4b_unlock(sc, 0);
					return (error);
				}
			}

			if (first) {
				sc->sc_daemon_chunk = *(struct v4b_data_chunk *)data;
				sc->sc_error_to_daemon = 0;
				first = 0;
			}
			error = sc->sc_error_to_daemon;
			if (error)
				break;

			/* check if complete */
			if (sc->sc_daemon_chunk.length == 0)
				break;

			temp = sc->sc_daemon_chunk.length;
			if (temp > V4B_BUFFER_MAX)
				temp = V4B_BUFFER_MAX;

			sc->sc_daemon_len = temp;
			sc->sc_cmd_from_daemon = V4B_CMD_READ;
			cv_signal(&sc->sc_cv);

			/* wait for ready state */
			while (sc->sc_cmd_from_daemon != V4B_CMD_NONE) {
				error = cv_wait_sig(&sc->sc_cv, &sc->sc_mtx);
				if (sc->sc_daemon_opened == 0) {
					v4b_unlock(sc, 0);
					return (ENXIO);
				}
				if (error) {
					v4b_unlock(sc, 0);
					return (error);
				}
			}

			/* transfer data */
			v4b_unlock(sc, 0);
#ifdef V4B_DEBUG
			printf("copyoutB(%p,%p,%d)\n",
			    sc->sc_buffer,
			    sc->sc_daemon_chunk.client_ptr,
			    temp);
#endif
			error = copyout(sc->sc_buffer, sc->sc_daemon_chunk.client_ptr, temp);
			v4b_lock(sc, 0);

			if (error)
				break;

			sc->sc_daemon_chunk.client_ptr += temp;
			sc->sc_daemon_chunk.peer_ptr += temp;
			sc->sc_daemon_chunk.length -= temp;
		}
		break;

	case V4B_IOCTL_ALLOC_MEMORY:

		mac = (struct v4b_mem_alloc *)data;
#ifdef V4B_DEBUG
		printf("alloc(%d,%d)\n", mac->alloc_nr, mac->page_count);
#endif
		if (mac->alloc_nr >= V4B_ALLOC_UNIT_MAX) {
			error = ENOMEM;
			break;
		}
		if (mac->page_count > V4B_ALLOC_PAGES_MAX) {
			error = ENOMEM;
			break;
		}
		error = v4b_alloc_memory(sc,
		    &sc->sc_mem[mac->alloc_nr],
		    mac->page_count);
		break;

	case V4B_IOCTL_FREE_MEMORY:

		mac = (struct v4b_mem_alloc *)data;
#ifdef V4B_DEBUG
		printf("free(%d,%d)\n", mac->alloc_nr, mac->page_count);
#endif
		if (mac->alloc_nr >= V4B_ALLOC_UNIT_MAX) {
			error = ENOMEM;
			break;
		}
		sc->sc_mem[mac->alloc_nr].in_use_by_daemon = 0;
		error = 0;
		break;

	case V4B_IOCTL_MAP_MEMORY:

		mac = (struct v4b_mem_alloc *)data;
#ifdef V4B_DEBUG
		printf("map(%d,%d)\n", mac->alloc_nr, mac->page_count);
#endif
		sc->sc_mmap_page_offset = mac->page_count;
		sc->sc_mmap_alloc_nr = mac->alloc_nr;
		break;

	default:
		error = ENOTTY;
		break;
	}

	v4b_unlock(sc, 0);

	return (error);
}

static int
v4b_daemon_poll(struct cdev *dev, int events, struct thread *td)
{
	return (events & (POLLHUP | POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM));
}

static int
#if (__FreeBSD_version >= 900000)
v4b_daemon_mmap(struct cdev *dev, vm_ooffset_t offset, vm_paddr_t *paddr, int nprot, vm_memattr_t *memattr)
#else
v4b_daemon_mmap(struct cdev *dev, vm_offset_t offset, vm_paddr_t *paddr, int nprot)
#endif
{
	struct v4b_softc *sc = (struct v4b_softc *)dev->si_drv1;
	uint32_t page_nr = offset / PAGE_SIZE;
	uint32_t alloc_nr = page_nr / V4B_ALLOC_PAGES_MAX;

	if (alloc_nr >= V4B_ALLOC_UNIT_MAX)
		return (ENOMEM);

	if (sc->sc_mem[alloc_nr].virtaddr == NULL)
		return (ENOMEM);

	page_nr %= V4B_ALLOC_PAGES_MAX;

	if (page_nr >= sc->sc_mem[alloc_nr].page_count)
		return (ENXIO);

	*paddr = vtophys(sc->sc_mem[alloc_nr].virtaddr +
	    (page_nr * PAGE_SIZE));

#ifdef V4B_DEBUG
	printf("mmap(%p,%d,%d)\n", (void *)*paddr, alloc_nr, page_nr);
#endif
	return (0);
}
